@@ -1,5 +1,3 @@
-# vim: tabstop=4 shiftwidth=4 softtabstop=4
-
 # Copyright 2015-2017 Lenovo
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -16,19 +14,17 @@
 
 import base64
 import binascii
+import socket
+import struct
 import traceback
-import urllib
+import weakref
 
 import pyghmi.constants as pygconst
 import pyghmi.exceptions as pygexc
 import pyghmi.ipmi.oem.generic as generic
-import pyghmi.ipmi.private.constants as ipmiconst
-import pyghmi.ipmi.private.util as util
-
 from pyghmi.ipmi.oem.lenovo import cpu
 from pyghmi.ipmi.oem.lenovo import dimm
 from pyghmi.ipmi.oem.lenovo import drive
-
 from pyghmi.ipmi.oem.lenovo import firmware
 from pyghmi.ipmi.oem.lenovo import imm
 from pyghmi.ipmi.oem.lenovo import inventory
@@ -37,13 +33,17 @@ from pyghmi.ipmi.oem.lenovo import pci
 from pyghmi.ipmi.oem.lenovo import psu
 from pyghmi.ipmi.oem.lenovo import raid_controller
 from pyghmi.ipmi.oem.lenovo import raid_drive
-
-
+import pyghmi.ipmi.private.constants as ipmiconst
+import pyghmi.ipmi.private.util as util
+import pyghmi.redfish.command as redfishcmd
+from pyghmi.redfish.oem.lenovo import tsma
 import pyghmi.util.webclient as wc
 
-import socket
-import struct
-import weakref
+try:
+    from urllib import urlencode
+except ImportError:
+    from urllib.parse import urlencode
+
 try:
     range = xrange
 except NameError:
@@ -156,6 +156,16 @@ class OEMHandler(generic.OEMHandler):
             self.immhandler = imm.IMMClient(ipmicmd)
         elif self.is_fpc:
             self.smmhandler = nextscale.SMMClient(ipmicmd)
+        elif self.has_tsma:
+            conn = wc.SecureHTTPConnection(
+                ipmicmd.bmc, 443,
+                verifycallback=self.ipmicmd.certverify)
+            self.tsmahandler = tsma.TsmHandler(None, None, conn,
+                                               fish=redfishcmd)
+            self.tsmahandler.set_credentials(
+                ipmicmd.ipmi_session.userid.decode('utf-8'),
+                ipmicmd.ipmi_session.password.decode('utf-8'))
+            self.tsmahandler.init_redfish()
 
     @property
     def _megarac_eth_index(self):
@@ -240,12 +250,12 @@ class OEMHandler(generic.OEMHandler):
                     event['severity'] = pygconst.Health.Ok
                     event['component'] = 'User Privilege'
                     event['component_type'] = ipmiconst.sensor_type_codes[6]
-                    event['event_data'] = \
-                        'User {0} on channel {1} had privilege changed ' \
+                    event['event_data'] = (
+                        'User {0} on channel {1} had privilege changed '
                         'from {2} to {3}'.format(
                             oemdata[2], oemdata[1], oemdata[3] & 0b1111,
-                            (oemdata[3] & 0b11110000) >> 4
-                        )
+                            (oemdata[3] & 0b11110000) >> 4)
+                    )
             else:
                 event['event'] = 'OEM event: {0}'.format(
                     ' '.join(format(x, '02x') for x in event['oemdata']))
@@ -326,10 +336,19 @@ class OEMHandler(generic.OEMHandler):
             return True
         return None
 
+    def set_user_access(self, uid, channel, callback, link_auth, ipmi_msg,
+                        privilege_level):
+        if self.is_fpc and self._fpc_variant == 2:
+            self.smmhandler.set_user_priv(uid, privilege_level)
+
     @property
     def is_fpc(self):
-        """True if the target is a Lenovo nextscale fan power controller
-        """
+        """True if the target is a Lenovo nextscale fan power controller"""
+
+        if self.has_imm or self.has_xcc:
+            return None
+        if self._fpc_variant is not None:
+            return self._fpc_variant
         fpc_ids = ((19046, 32, 1063), (20301, 32, 462))
         smm_id = (19046, 32, 1180)
         currid = (self.oemid['manufacturer_id'], self.oemid['device_id'],
@@ -346,10 +365,17 @@ class OEMHandler(generic.OEMHandler):
                                       self.oemid['device_id'],
                                       self.oemid['product_id'])
 
+    tsma_ids = ((19046, 32, 1287),)
+    @property
+    def has_tsma(self):
+        currid = (self.oemid['manufacturer_id'], self.oemid['device_id'],
+                  self.oemid['product_id'])
+        return currid in self.tsma_ids
+
     @property
     def has_tsm(self):
-        """True if this particular server have a TSM based service processor
-        """
+        """True if this particular server have a TSM based service processor"""
+
         if (self.oemid['manufacturer_id'] == 19046 and
                 self.oemid['device_id'] == 32):
             try:
@@ -381,29 +407,29 @@ class OEMHandler(generic.OEMHandler):
                 yield inv
 
     def get_sensor_data(self):
-        if self.is_fpc:
-            for name in nextscale.get_sensor_names(self._fpc_variant):
-                yield nextscale.get_sensor_reading(name, self.ipmicmd,
-                                                   self._fpc_variant)
-        elif self.has_imm:
+        if self.has_imm:
             for name in self.immhandler.get_oem_sensor_names(self.ipmicmd):
                 yield self.immhandler.get_oem_sensor_reading(name,
                                                              self.ipmicmd)
+        elif self.is_fpc:
+            for name in nextscale.get_sensor_names(self._fpc_variant):
+                yield nextscale.get_sensor_reading(name, self.ipmicmd,
+                                                   self._fpc_variant)
 
     def get_sensor_descriptions(self):
-        if self.is_fpc:
-            return nextscale.get_sensor_descriptions(self._fpc_variant)
-        elif self.has_imm:
+        if self.has_imm:
             return self.immhandler.get_oem_sensor_descriptions(self.ipmicmd)
+        elif self.is_fpc:
+            return nextscale.get_sensor_descriptions(self._fpc_variant)
         return ()
 
     def get_sensor_reading(self, sensorname):
-        if self.is_fpc:
-            return nextscale.get_sensor_reading(sensorname, self.ipmicmd,
-                                                self._fpc_variant)
-        elif self.has_imm:
+        if self.has_imm:
             return self.immhandler.get_oem_sensor_reading(sensorname,
                                                           self.ipmicmd)
+        elif self.is_fpc:
+            return nextscale.get_sensor_reading(sensorname, self.ipmicmd,
+                                                self._fpc_variant)
         return ()
 
     def get_inventory_of_component(self, component):
@@ -483,7 +509,7 @@ class OEMHandler(generic.OEMHandler):
         else:
             raise pygexc.UnsupportedFunctionality()
 
-    def process_fru(self, fru):
+    def process_fru(self, fru, name=None):
         if fru is None:
             return fru
         if self.has_tsm:
@@ -529,7 +555,7 @@ class OEMHandler(generic.OEMHandler):
                 endidx = len(macs) - 5
                 macprefix = None
                 while idx < endidx:
-                    currmac = macs[idx:idx+6]
+                    currmac = macs[idx:idx + 6]
                     if not isinstance(currmac, bytearray):
                         # invalid vpd format, abort attempts to extract
                         # mac in this way
@@ -547,6 +573,8 @@ class OEMHandler(generic.OEMHandler):
                     idx = idx + 6
             except (AttributeError, KeyError, IndexError):
                 pass
+            if self.has_xcc and name and name.startswith('PSU '):
+                self.immhandler.augment_psu_info(fru, name)
             return fru
         elif self.is_fpc == 2:  # SMM variant
             fru['oem_parser'] = 'lenovo'
@@ -605,7 +633,21 @@ class OEMHandler(generic.OEMHandler):
         elif self.is_fpc:
             return nextscale.get_fpc_firmware(bmcver, self.ipmicmd,
                                               self._fpc_variant)
-        return super(OEMHandler, self).get_oem_firmware(bmcver)
+        elif self.has_tsma:
+            return self.tsmahandler.get_firmware_inventory(components,
+                                                           raisebypass=False)
+        return super(OEMHandler, self).get_oem_firmware(bmcver, components)
+
+    def get_diagnostic_data(self, savefile, progress, autosuffix=False):
+        if self.has_xcc:
+            return self.immhandler.get_diagnostic_data(savefile, progress,
+                                                       autosuffix)
+        if self.is_fpc:
+            return self.smmhandler.get_diagnostic_data(savefile, progress,
+                                                       autosuffix)
+        if self.has_tsma:
+            return self.tsmahandler.get_diagnostic_data(savefile, progress,
+                                                        autosuffix)
 
     def get_oem_capping_enabled(self):
         if self.has_tsm:
@@ -664,9 +706,9 @@ class OEMHandler(generic.OEMHandler):
             # set the domain name content
             name = name.ljust(256, "\x00")
             for i in range(0, 4):
-                data = [4, i+1]
-                offset = i*64
-                data.extend([ord(x) for x in name[offset:offset+64]])
+                data = [4, i + 1]
+                offset = i * 64
+                data.extend([ord(x) for x in name[offset:offset + 64]])
                 self.ipmicmd.xraw_command(netfn=0x32, command=0x6c, data=data)
 
             self._restart_dns()
@@ -675,17 +717,17 @@ class OEMHandler(generic.OEMHandler):
             self.smmhandler.set_domain(name)
 
     def set_hostname(self, hostname):
-        if self.is_fpc:
-            return self.smmhandler.set_hostname(hostname)
-        elif self.has_xcc:
+        if self.has_xcc:
             return self.immhandler.set_hostname(hostname)
+        elif self.is_fpc:
+            return self.smmhandler.set_hostname(hostname)
         return super(OEMHandler, self).set_hostname(hostname)
 
     def get_hostname(self):
-        if self.is_fpc:
-            return self.smmhandler.get_hostname()
-        elif self.has_xcc:
+        if self.has_xcc:
             return self.immhandler.get_hostname()
+        elif self.is_fpc:
+            return self.smmhandler.get_hostname()
         return super(OEMHandler, self).get_hostname()
 
     """ Gets a remote console launcher for a Lenovo ThinkServer.
@@ -700,7 +742,7 @@ class OEMHandler(generic.OEMHandler):
         conn = wc.SecureHTTPConnection(bmc, 443,
                                        verifycallback=self.ipmicmd.certverify)
         conn.connect()
-        params = urllib.urlencode({
+        params = urlencode({
             'WEBVAR_USERNAME': username,
             'WEBVAR_PASSWORD': password
         })
@@ -757,7 +799,7 @@ class OEMHandler(generic.OEMHandler):
             else:
                 # fall back to a dumber, but more universal formatter
                 ipv6str = binascii.b2a_hex(ipv6_addr)
-                ipv6str = ':'.join([ipv6str[x:x+4] for x in range(0, 32, 4)])
+                ipv6str = ':'.join([ipv6str[x:x + 4] for x in range(0, 32, 4)])
             netdata['ipv6_addresses'] = [
                 '{0}/{1}'.format(ipv6str, ipv6_prefix)]
 
@@ -816,7 +858,7 @@ class OEMHandler(generic.OEMHandler):
                                   data=(1, selector, 0, 1))
         # now do the set
         for x in range(0, 256, 64):
-            currdata = padded[x:x+64]
+            currdata = padded[x:x + 64]
             currchunk = x // 64 + 1
             cmddata = [1, selector, currchunk] + currdata
             self.ipmicmd.xraw_command(netfn=0x32, command=0x9f, data=cmddata)
@@ -830,7 +872,7 @@ class OEMHandler(generic.OEMHandler):
         imgnames = rsp['data'][1:]
         shortnames = []
         for idx in range(0, len(imgnames), 22):
-            shortnames.append(imgnames[idx+2:idx+22].rstrip('\0'))
+            shortnames.append(imgnames[idx + 2:idx + 22].rstrip('\0'))
         return shortnames
 
     def _megarac_media_waitforready(self, imagename):
@@ -882,6 +924,9 @@ class OEMHandler(generic.OEMHandler):
     def attach_remote_media(self, url, username, password):
         if self.has_imm:
             self.immhandler.attach_remote_media(url, username, password)
+        elif self.has_tsma:
+            return self.tsmahandler.attach_remote_media(
+                url, username, password, None)
         elif self.has_megarac:
             proto, host, path = util.urlsplit(url)
             if proto == 'smb':
@@ -910,6 +955,9 @@ class OEMHandler(generic.OEMHandler):
         if self.is_fpc:
             return self.smmhandler.update_firmware(
                 filename, data=data, progress=progress, bank=bank)
+        if self.has_tsma:
+            return self.tsmahandler.update_firmware(
+                filename, data=data, progress=progress, bank=bank)
         return super(OEMHandler, self).update_firmware(filename, data=data,
                                                        progress=progress,
                                                        bank=bank)
@@ -917,21 +965,62 @@ class OEMHandler(generic.OEMHandler):
     def get_description(self):
         if self.has_xcc:
             return self.immhandler.get_description()
+        if self.is_fpc:
+            return {'height': self._fpc_variant, 'slot': 0}
         return super(OEMHandler, self).get_description()
 
+    def get_extended_bmc_configuration(self):
+        if self.has_xcc:
+            return self.immhandler.get_extended_bmc_configuration()
+        return super(OEMHandler, self).get_extended_bmc_configuration()
+
+    def get_bmc_configuration(self):
+        if self.has_xcc:
+            return self.immhandler.get_bmc_configuration()
+        if self.is_fpc:
+            return self.smmhandler.get_bmc_configuration()
+        return super(OEMHandler, self).get_bmc_configuration()
+
+    def set_bmc_configuration(self, changeset):
+        if self.has_xcc:
+            return self.immhandler.set_bmc_configuration(changeset)
+        if self.is_fpc:
+            return self.smmhandler.set_bmc_configuration(changeset)
+        return super(OEMHandler, self).set_bmc_configuration(changeset)
+
     def get_system_configuration(self, hideadvanced):
         if self.has_imm or self.has_xcc:
             return self.immhandler.get_system_configuration(hideadvanced)
+        if self.has_tsma:
+            return self.tsmahandler.get_uefi_configuration(hideadvanced)
         return super(OEMHandler, self).get_system_configuration(hideadvanced)
 
     def set_system_configuration(self, changeset):
         if self.has_imm or self.has_xcc:
             return self.immhandler.set_system_configuration(changeset)
+        if self.has_tsma:
+            return self.tsmahandler.set_uefi_configuration(changeset)
         return super(OEMHandler, self).set_system_configuration(changeset)
 
+    def clear_bmc_configuration(self):
+        if self.has_xcc:
+            return self.immhandler.clear_bmc_configuration()
+        elif self.is_fpc:
+            return self.smmhandler.clear_bmc_configuration()
+        return super(OEMHandler, self).clear_system_configuration()
+
+    def clear_system_configuration(self):
+        if self.has_xcc:
+            return self.immhandler.clear_system_configuration()
+        if self.has_tsma:
+            return self.tsmahandler.clear_uefi_configuration()
+        return super(OEMHandler, self).clear_system_configuration()
+
     def detach_remote_media(self):
         if self.has_imm:
             self.immhandler.detach_remote_media()
+        elif self.has_tsma:
+            self.tsmahandler.detach_remote_media()
         elif self.has_megarac:
             self.ipmicmd.xraw_command(
                 netfn=0x32, command=0x9f, data=(8, 10, 0, 0))
@@ -945,9 +1034,36 @@ class OEMHandler(generic.OEMHandler):
     def list_media(self):
         if self.has_xcc or self.has_imm:
             return self.immhandler.list_media()
+        if self.has_tsma:
+            return self.tsmahandler.list_media()
         return super(OEMHandler, self).list_media()
 
     def get_health(self, summary):
         if self.has_xcc:
             return self.immhandler.get_health(summary)
         return super(OEMHandler, self).get_health(summary)
+
+    def get_licenses(self):
+        if self.has_xcc:
+            return self.immhandler.get_licenses()
+        return super(OEMHandler, self).get_licenses()
+
+    def get_user_expiration(self, uid):
+        if self.has_xcc:
+            return self.immhandler.get_user_expiration(uid)
+        return None
+
+    def delete_license(self, name):
+        if self.has_xcc:
+            return self.immhandler.delete_license(name)
+        return super(OEMHandler, self).delete_license(name)
+
+    def save_licenses(self, directory):
+        if self.has_xcc:
+            return self.immhandler.save_licenses(directory)
+        return super(OEMHandler, self).save_licenses(directory)
+
+    def apply_license(self, filename, progress=None):
+        if self.has_xcc:
+            return self.immhandler.apply_license(filename, progress)
+        return super(OEMHandler, self).apply_license(filename, progress)
