@@ -1,5 +1,3 @@
-# vim: tabstop=4 shiftwidth=4 softtabstop=4
-
 # Copyright 2013 IBM Corporation
 # Copyright 2015-2017 Lenovo
 #
@@ -14,7 +12,8 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-# This represents the low layer message framing portion of IPMI
+
+"""This represents the low layer message framing portion of IPMI"""
 
 import collections
 import hashlib
@@ -27,14 +26,15 @@ import socket
 import struct
 import threading
 
-
 from cryptography.hazmat.backends import default_backend
-from cryptography.hazmat.primitives.ciphers import algorithms, Cipher, modes
-
+from cryptography.hazmat.primitives.ciphers import algorithms
+from cryptography.hazmat.primitives.ciphers import Cipher
+from cryptography.hazmat.primitives.ciphers import modes
 import pyghmi.exceptions as exc
 from pyghmi.ipmi.private import constants
 from pyghmi.ipmi.private import util
-from pyghmi.ipmi.private.util import get_ipmi_error, _monotonic_time
+from pyghmi.ipmi.private.util import _monotonic_time
+from pyghmi.ipmi.private.util import get_ipmi_error
 
 
 KEEPALIVE_SESSIONS = threading.RLock()
@@ -107,7 +107,7 @@ def define_worker():
                 if timeout < 0:
                     timeout = 0
                 selectdeadline = _monotonic_time() + timeout
-                tmplist, _, _ = select.select(iosockets, (), (), timeout)
+                select.select(iosockets, (), (), timeout)
                 # pessimistically move out the deadline
                 # doing it this early (before ioqueue is evaluated)
                 # this avoids other threads making a bad assumption
@@ -250,8 +250,9 @@ def _poller(timeout=0):
 
 
 def _aespad(data):
-    """ipmi demands a certain pad scheme,
-    per table 13-20 AES-CBC encrypted payload fields.
+    """ipmi demands a certain pad scheme, per table 13-20 AES-CBC encrypted
+
+    payload fields.
     """
     currlen = len(data) + 1  # need to count the pad length field as well
     neededpad = currlen % 16
@@ -390,6 +391,7 @@ class Session(object):
             iothreadwaiters.append(initevt)
             _IOWorker = define_worker()
             iothread = _IOWorker()
+            iothread.daemon = True
             iothread.start()
             initevt.wait()
         elif not iothreadready:
@@ -399,9 +401,7 @@ class Session(object):
         return tmpsocket
 
     def _sync_login(self, response):
-        """Handle synchronous callers in liue of
-        a client-provided callback.
-        """
+        """Handle synchronous callers in lieu of a client-provided callback"""
         # Be a stub, the __init__ will catch and respond to ensure response
         # is given in the same thread as was called
         return
@@ -423,7 +423,8 @@ class Session(object):
                 port=623,
                 kg=None,
                 onlogon=None,
-                privlevel=4):
+                privlevel=None,
+                keepalive=True):
         trueself = None
         forbidsock = []
         for res in socket.getaddrinfo(bmc, port, 0, socket.SOCK_DGRAM):
@@ -471,7 +472,8 @@ class Session(object):
                  port=623,
                  kg=None,
                  onlogon=None,
-                 privlevel=4):
+                 privlevel=None,
+                 keepalive=True):
         if hasattr(self, 'initialized'):
             # new found an existing session, do not corrupt it
             if onlogon is None:
@@ -484,8 +486,16 @@ class Session(object):
                     self.iterwaiters.append(onlogon)
             return
         self.broken = False
+        self.socket = None
         self.logged = 0
-        self.privlevel = privlevel
+        if privlevel is not None:
+            self.privlevel = privlevel
+            self.autopriv = False
+        else:
+            self.privlevel = 4
+            self.autopriv = True
+        self.logoutexpiry = None
+        self.autokeepalive = keepalive
         self.maxtimeout = 3  # be aggressive about giving up on initial packet
         self.incommand = False
         self.nameonly = 16  # default to name only lookups in RAKP exchange
@@ -540,6 +550,7 @@ class Session(object):
     def _mark_broken(self, error=None):
         # since our connection has failed retries
         # deregister our keepalive facility
+        self.lastpayload = None
         with util.protect(KEEPALIVE_SESSIONS):
             Session.keepalive_sessions.pop(self, None)
         with util.protect(WAITING_SESSIONS):
@@ -550,12 +561,13 @@ class Session(object):
                                            self.kgo)]
         except KeyError:
             pass
-        self.logout()
+        self.logout(False)
         self.logging = False
         self.errormsg = error
         if not self.broken:
             self.broken = True
-            self.socketpool[self.socket] -= 1
+            if self.socket:
+                self.socketpool[self.socket] -= 1
 
     def onlogon(self, parameter):
         if 'error' in parameter:
@@ -572,6 +584,9 @@ class Session(object):
         self.confalgo = 0
         self.aeskey = None
         self.integrityalgo = 0
+        self.attemptedhash = 256
+        self.currhashlib = None
+        self.currhashlen = 0
         self.k1 = None
         self.rmcptag = 1
         self.lastpayload = None
@@ -582,6 +597,8 @@ class Session(object):
         self.authtype = 0
         self.ipmiversion = 1.5
         self.timeout = initialtimeout + (0.5 * random.random())
+        self.logoutexpiry = _monotonic_time() + self._getmaxtimeout()
+        self.rqlun = 0
         self.seqlun = 0
         # NOTE(jbjohnso): per IPMI table 5-4, software ids in the ipmi spec may
         #                 be 0x81 through 0x8d.  We'll stick with 0x81 for now,
@@ -607,14 +624,17 @@ class Session(object):
 
     def _make_bridge_request_msg(self, channel, netfn, command):
         """This function generate message for bridge request. It is a
+
         part of ipmi payload.
         """
+        # NOTE(puwen): need to pay attention for this function because the
+        # structure does not seem to match the specifications.
         head = bytearray((constants.IPMI_BMC_ADDRESS,
                           constants.netfn_codes['application'] << 2))
         check_sum = _checksum(*head)
         # NOTE(fengqian): according IPMI Figure 14-11, rqSWID is set to 81h
-        boday = bytearray((0x81, self.seqlun, constants.IPMI_SEND_MESSAGE_CMD,
-                           0x40 | channel))
+        boday = bytearray((0x81, (self.seqlun << 2) | self.rqlun,
+                           constants.IPMI_SEND_MESSAGE_CMD, 0x40 | channel))
         # NOTE(fengqian): Track request
         self._add_request_entry((constants.netfn_codes['application'] + 1,
                                  self.seqlun, constants.IPMI_SEND_MESSAGE_CMD))
@@ -622,6 +642,7 @@ class Session(object):
 
     def _add_request_entry(self, entry=()):
         """This function record the request with netfn, sequence number and
+
         command, which will be used in parse_ipmi_payload.
         :param entry: a set of netfn, sequence number and command.
         """
@@ -635,8 +656,10 @@ class Session(object):
         if self._lookup_request_entry(entry):
             self.request_entry.remove(entry)
 
-    def _make_ipmi_payload(self, netfn, command, bridge_request=None, data=()):
+    def _make_ipmi_payload(self, netfn, command, bridge_request=None, data=(),
+                           rslun=0):
         """This function generates the core ipmi payload that would be
+
         applicable for any channel (including KCS)
         """
         bridge_msg = []
@@ -652,8 +675,8 @@ class Session(object):
                self.tabooseq[(netfn, command, self.seqlun)] and seqincrement):
             self.tabooseq[(self.expectednetfn, command, self.seqlun)] -= 1
             # Allow taboo to eventually expire after a few rounds
-            self.seqlun += 4  # the last two bits are lun, so add 4 to add 1
-            self.seqlun &= 0xff  # we only have one byte, wrap when exceeded
+            self.seqlun += 1  # the last two bits are lun, so add 4 to add 1
+            self.seqlun &= 0x3f  # we only have one byte, wrap when exceeded
             seqincrement -= 1
 
         if bridge_request:
@@ -672,9 +695,9 @@ class Session(object):
         # figure 13-4, first two bytes are rsaddr and
         # netfn, for non-bridge request, rsaddr is always 0x20 since we are
         # addressing BMC while rsaddr is specified forbridge request
-        header = bytearray((rsaddr, netfn << 2))
-
-        reqbody = bytearray((rqaddr, self.seqlun, command)) + data
+        header = bytearray((rsaddr, (netfn << 2) | rslun))
+        reqbody = bytearray(
+            (rqaddr, (self.seqlun << 2) | self.rqlun, command)) + data
         headsum = bytearray((_checksum(*header),))
         bodysum = bytearray((_checksum(*reqbody),))
         payload = header + headsum + reqbody + bodysum
@@ -699,6 +722,8 @@ class Session(object):
             stillin = self.incommand - _monotonic_time()
             if stillin > 0:
                 return stillin
+            else:
+                self.lastpayload = None
         return 0
 
     def _getmaxtimeout(self):
@@ -735,8 +760,12 @@ class Session(object):
                     retry=True,
                     delay_xmit=None,
                     timeout=None,
-                    callback=None):
+                    callback=None,
+                    rslun=0):
         if not self.logged:
+            if (self.logoutexpiry is not None and
+                    _monotonic_time() > self.logoutexpiry):
+                self._mark_broken()
             raise exc.IpmiException('Session no longer connected')
         self._cmdwait()
         if not self.logged:
@@ -750,7 +779,7 @@ class Session(object):
         self._send_ipmi_net_payload(netfn, command, data,
                                     bridge_request=bridge_request,
                                     retry=retry, delay_xmit=delay_xmit,
-                                    timeout=timeout)
+                                    timeout=timeout, rslun=rslun)
 
         if retry:  # in retry case, let the retry timers indicate wait time
             timeout = None
@@ -776,7 +805,8 @@ class Session(object):
 
     def _send_ipmi_net_payload(self, netfn=None, command=None, data=(), code=0,
                                bridge_request=None,
-                               retry=None, delay_xmit=None, timeout=None):
+                               retry=None, delay_xmit=None, timeout=None,
+                               rslun=0):
         if retry is None:
             retry = not self.servermode
         if self.servermode:
@@ -788,7 +818,7 @@ class Session(object):
         else:
             data = bytearray(data)
         ipmipayload = self._make_ipmi_payload(netfn, command, bridge_request,
-                                              data)
+                                              data, rslun)
         payload_type = constants.payload_types['ipmi']
         self.send_payload(payload=ipmipayload, payload_type=payload_type,
                           retry=retry, delay_xmit=delay_xmit, timeout=timeout)
@@ -890,9 +920,9 @@ class Session(object):
                 message.append(neededpad)
                 message.append(7)  # reserved, 7 is the required value for the
                 # specification followed
-                message += hmac.new(self.k1,
-                                    bytes(message[4:]),
-                                    hashlib.sha1).digest()[:12]  # SHA1-96
+                message += hmac.new(
+                    self.k1, bytes(message[4:]),
+                    self.currhashlib).digest()[:self.currhashlen]
                 # per RFC2404 truncates to 96 bits
         self.netpacket = message
         # advance idle timer since we don't need keepalive while sending
@@ -917,7 +947,7 @@ class Session(object):
             raise exc.IpmiException("Password is too long for ipmi 1.5")
         password += '\x00' * padneeded
         if checkremotecode:
-            seqbytes = struct.pack("<I", self.remsequencenumber)
+            seqbytes = struct.pack("<I", self.remseqnumber)
         else:
             seqbytes = struct.pack("<I", self.sequencenumber)
         sessdata = struct.pack("<I", self.sessionid)
@@ -961,7 +991,7 @@ class Session(object):
             self.onlogon({'error': errstr})
             return
         data = response['data']
-        self.sessionid = struct.unpack("<I", struct.pack("4B", *data[0:4]))[0]
+        self.sessionid = struct.unpack("<I", bytes(data[0:4]))[0]
         self.authtype = 2
         self._activate_session(data[4:])
 
@@ -981,13 +1011,13 @@ class Session(object):
             self.onlogon({'error': errstr})
             return
         data = response['data']
-        self.sessionid = struct.unpack("<I", struct.pack("4B", *data[1:5]))[0]
-        self.sequencenumber = struct.unpack("<I",
-                                            struct.pack("4B", *data[5:9]))[0]
+        self.sessionid = struct.unpack("<I", bytes(data[1:5]))[0]
+        self.sequencenumber = struct.unpack("<I", bytes(data[5:9]))[0]
         self._req_priv_level()
 
     def _req_priv_level(self):
         self.logged = 1
+        self.logoutexpiry = None
         response = self.raw_command(netfn=0x6, command=0x3b,
                                     data=[self.privlevel])
         if response['code']:
@@ -1015,13 +1045,13 @@ class Session(object):
         self.onlogon({'success': True})
 
     def _get_session_challenge(self):
-        reqdata = [2]
+        reqdata = bytearray([2])
         if len(self.userid) > 16:
             raise exc.IpmiException(
                 "Username too long for IPMI, must not exceed 16")
         padneeded = 16 - len(self.userid)
         userid = self.userid + ('\x00' * padneeded)
-        reqdata += struct.unpack("!16B", userid)
+        reqdata += userid
         self.ipmicallback = self._got_session_challenge
         self._send_ipmi_net_payload(netfn=0x6, command=0x39, data=reqdata)
 
@@ -1031,19 +1061,34 @@ class Session(object):
         # login attempts from the past
         self.localsid += 1
         self.rmcptag += 1
-        data = [
+        data = bytearray([
             self.rmcptag,
             0,  # request as much privilege as the channel will give us
             0, 0,  # reserved
-        ]
-        data += list(struct.unpack("4B", struct.pack("<I", self.localsid)))
-        data += [
-            0, 0, 0, 8, 1, 0, 0, 0,  # table 13-17, SHA-1
-            1, 0, 0, 8, 1, 0, 0, 0,  # SHA-1 integrity
-            2, 0, 0, 8, 1, 0, 0, 0,  # AES privacy
-            # 2,0,0,8,0,0,0,0, #no privacy confalgo
-        ]
+        ])
+        data += struct.pack("<I", self.localsid)
+        # auth 3 sha256
+        # integrity... 4 = sha256
+        if self.attemptedhash == 1:
+            data += bytearray([
+                0, 0, 0, 8, 1, 0, 0, 0,  # table 13-17, SHA-1
+                1, 0, 0, 8, 1, 0, 0, 0,  # SHA-1 integrity
+                2, 0, 0, 8, 1, 0, 0, 0,  # AES privacy
+                # 2,0,0,8,0,0,0,0, #no privacy confalgo
+            ])
+            self.currhashlib = hashlib.sha1
+            self.currhashlen = 12
+        else:
+            data += bytearray([
+                0, 0, 0, 8, 3, 0, 0, 0,  # table 13-17, SHA-256
+                1, 0, 0, 8, 4, 0, 0, 0,  # SHA-256-128 integrity
+                2, 0, 0, 8, 1, 0, 0, 0,  # AES privacy
+                # 2,0,0,8,0,0,0,0, #no privacy confalgo
+            ])
+            self.currhashlib = hashlib.sha256
+            self.currhashlen = 16
         self.sessioncontext = 'OPENSESSION'
+        self.lastpayload = None
         self.send_payload(
             payload=data,
             payload_type=constants.payload_types['rmcpplusopenreq'])
@@ -1209,8 +1254,8 @@ class Session(object):
         return _keptalive
 
     def _keepalive(self):
-        """Performs a keepalive to avoid idle disconnect
-        """
+        """Performs a keepalive to avoid idle disconnect"""
+
         try:
             keptalive = False
             if self._customkeepalives:
@@ -1235,8 +1280,11 @@ class Session(object):
                 if self.incommand:
                     # if currently in command, no cause to keepalive
                     return
-                self.raw_command(netfn=6, command=1,
-                                 callback=self._keepalive_wrapper(None))
+                if self.autokeepalive:
+                    self.raw_command(netfn=6, command=1,
+                                     callback=self._keepalive_wrapper(None))
+                else:
+                    self.logout()
         except exc.IpmiException:
             self._mark_broken()
 
@@ -1264,9 +1312,9 @@ class Session(object):
             # things off ignore the second reply since we have one
             # satisfactory answer
         if data[4] in (0, 2):  # This is an ipmi 1.5 paylod
-            remsequencenumber = struct.unpack('<I', data[5:9])[0]
-            remsessid = struct.unpack("<I", data[9:13])[0]
-            if (remsequencenumber == 0 and remsessid == 0 and
+            remseqnumber = struct.unpack('<I', bytes(data[5:9]))[0]
+            remsessid = struct.unpack("<I", bytes(data[9:13]))[0]
+            if (remseqnumber == 0 and remsessid == 0 and
                     qent[2] in Session.bmc_handlers):
                 # So a new ipmi client happens to get a previously seen and
                 # still active UDP source port.  Clear ourselves out and punt
@@ -1276,10 +1324,10 @@ class Session(object):
                 iserver.pktqueue.append(qent)
                 iserver.process_pktqueue()
                 return
-            if (hasattr(self, 'remsequencenumber') and
-                    remsequencenumber < self.remsequencenumber):
+            if (hasattr(self, 'remseqnumber') and
+                    remseqnumber < self.remseqnumber):
                 return -5  # remote sequence number is too low, reject it
-            self.remsequencenumber = remsequencenumber
+            self.remseqnumber = remseqnumber
             if data[4] != self.authtype:
                 # BMC responded with mismatch authtype, for
                 # mutual authentication reject it. If this causes
@@ -1342,17 +1390,18 @@ class Session(object):
             encrypted = 0
             if data[5] & 0b10000000:
                 encrypted = 1
-            authcode = data[-12:]
+            authcode = data[-self.currhashlen:]
             if self.k1 is None:  # we are in no shape to process a packet now
                 return
             expectedauthcode = hmac.new(
-                self.k1, data[4:-12], hashlib.sha1).digest()[:12]
+                self.k1, data[4:-self.currhashlen],
+                self.currhashlib).digest()[:self.currhashlen]
             if authcode != expectedauthcode:
                 return  # BMC failed to assure integrity to us, drop it
-            sid = struct.unpack("<I", data[6:10])[0]
+            sid = struct.unpack("<I", bytes(data[6:10]))[0]
             if sid != self.localsid:  # session id mismatch, drop it
                 return
-            remseqnumber = struct.unpack("<I", data[10:14])[0]
+            remseqnumber = struct.unpack("<I", bytes(data[10:14]))[0]
             if (hasattr(self, 'remseqnumber') and
                 (remseqnumber < self.remseqnumber) and
                     (self.remseqnumber != 0xffffffff)):
@@ -1403,6 +1452,10 @@ class Session(object):
         if data[0] != self.rmcptag:
             return -9  # use rmcp tag to track and reject stale responses
         if data[1] != 0:  # response code...
+            if self.attemptedhash == 256:
+                self.attemptedhash = 1
+                self._open_rmcpplus_request()
+                return
             if data[1] in constants.rmcp_codes:
                 errstr = constants.rmcp_codes[data[1]]
             else:
@@ -1414,11 +1467,10 @@ class Session(object):
         # shall be used.  As such, the allowedpriv field is actually
         # not particularly useful.  got_rakp2 is a good place to
         # gracefully detect and downgrade privilege for retry
-        localsid = struct.unpack("<I", struct.pack("4B", *data[4:8]))[0]
+        localsid = struct.unpack("<I", bytes(data[4:8]))[0]
         if self.localsid != localsid:
             return -9
-        self.pendingsessionid = struct.unpack(
-            "<I", struct.pack("4B", *data[8:12]))[0]
+        self.pendingsessionid = struct.unpack("<I", bytes(data[8:12]))[0]
         # TODO(jbjohnso): currently, we take it for granted that the responder
         # accepted our integrity/auth/confidentiality proposal
         self.lastpayload = None
@@ -1428,13 +1480,11 @@ class Session(object):
         self.rmcptag += 1
         self.randombytes = os.urandom(16)
         userlen = len(self.userid)
-        payload = [self.rmcptag, 0, 0, 0] + \
-            list(struct.unpack("4B",
-                 struct.pack("<I", self.pendingsessionid))) +\
-            list(struct.unpack("16B", self.randombytes)) +\
-            [self.nameonly | self.privlevel, 0, 0] +\
-            [userlen] +\
-            list(struct.unpack("%dB" % userlen, self.userid))
+        payload = bytearray([self.rmcptag, 0, 0, 0]) + \
+            struct.pack("<I", self.pendingsessionid) + \
+            self.randombytes +\
+            bytearray([self.nameonly | self.privlevel, 0, 0, userlen]) + \
+            self.userid
         self.sessioncontext = "EXPECTINGRAKP2"
         self.send_payload(
             payload=payload, payload_type=constants.payload_types['rakp1'])
@@ -1448,9 +1498,9 @@ class Session(object):
         if data[0] != self.rmcptag:  # ignore mismatched tags for retry logic
             return -9
         if data[1] != 0:  # if not successful, consider next move
-            if data[1] in (9, 0xd) and self.privlevel == 4:
+            if data[1] in (9, 0xd) and self.privlevel == 4 and self.autopriv:
                 # Here the situation is likely that the peer didn't want
-                # us to use Operator.  Degrade to operator and try again
+                # us to use admin.  Degrade to operator and try again
                 self.privlevel = 3
                 self.login()
                 return
@@ -1464,17 +1514,19 @@ class Session(object):
                 errstr = "Unrecognized RMCP code %d" % data[1]
             self.onlogon({'error': errstr + " in RAKP2"})
             return -9
-        localsid = struct.unpack("<I", struct.pack("4B", *data[4:8]))[0]
+        localsid = struct.unpack("<I", bytes(data[4:8]))[0]
         if localsid != self.localsid:
             return -9  # discard mismatch in the session identifier
-        self.remoterandombytes = struct.pack("16B", *data[8:24])
-        self.remoteguid = struct.pack("16B", *data[24:40])
+        self.remoterandombytes = bytes(data[8:24])
+        self.remoteguid = bytes(data[24:40])
         userlen = len(self.userid)
-        hmacdata = struct.pack("<II", localsid, self.pendingsessionid) +\
-            self.randombytes + self.remoterandombytes + self.remoteguid +\
-            struct.pack("2B", self.nameonly | self.privlevel, userlen) +\
-            self.userid
-        expectedhash = hmac.new(self.password, hmacdata, hashlib.sha1).digest()
+        hmacdata = (struct.pack("<II", localsid, self.pendingsessionid) +
+                    self.randombytes + self.remoterandombytes +
+                    self.remoteguid + struct.pack(
+                    "2B", self.nameonly | self.privlevel, userlen) +
+                    self.userid)
+        expectedhash = hmac.new(self.password, hmacdata,
+                                self.currhashlib).digest()
         hashlen = len(expectedhash)
         givenhash = struct.pack("%dB" % hashlen, *data[40:hashlen + 40])
         if givenhash != expectedhash:
@@ -1487,9 +1539,9 @@ class Session(object):
                             self.randombytes + self.remoterandombytes +
                             struct.pack("2B", self.nameonly | self.privlevel,
                                         userlen) +
-                            self.userid, hashlib.sha1).digest()
-        self.k1 = hmac.new(self.sik, b'\x01' * 20, hashlib.sha1).digest()
-        self.k2 = hmac.new(self.sik, b'\x02' * 20, hashlib.sha1).digest()
+                            self.userid, self.currhashlib).digest()
+        self.k1 = hmac.new(self.sik, b'\x01' * 20, self.currhashlib).digest()
+        self.k2 = hmac.new(self.sik, b'\x02' * 20, self.currhashlib).digest()
         self.aeskey = self.k2[0:16]
         self.sessioncontext = "EXPECTINGRAKP4"
         self.lastpayload = None
@@ -1506,7 +1558,7 @@ class Session(object):
                         len(self.userid)) +\
             self.userid
 
-        authcode = hmac.new(self.password, hmacdata, hashlib.sha1).digest()
+        authcode = hmac.new(self.password, hmacdata, self.currhashlib).digest()
         payload += list(struct.unpack("%dB" % len(authcode), authcode))
         self.send_payload(
             payload=payload, payload_type=constants.payload_types['rakp3'])
@@ -1535,21 +1587,21 @@ class Session(object):
                 errstr = "Unrecognized RMCP code %d" % data[1]
             self.onlogon({'error': errstr + " reported in RAKP4"})
             return -9
-        localsid = struct.unpack("<I", struct.pack("4B", *data[4:8]))[0]
+        localsid = struct.unpack("<I", bytes(data[4:8]))[0]
         if localsid != self.localsid:  # ignore if wrong session id indicated
             return -9
         hmacdata = self.randombytes +\
             struct.pack("<I", self.pendingsessionid) +\
             self.remoteguid
-        expectedauthcode = hmac.new(self.sik, hmacdata,
-                                    hashlib.sha1).digest()[:12]
+        expectedauthcode = hmac.new(
+            self.sik, hmacdata, self.currhashlib).digest()[:self.currhashlen]
         aclen = len(expectedauthcode)
         authcode = struct.pack("%dB" % aclen, *data[8:aclen + 8])
         if authcode != expectedauthcode:
             self.onlogon({'error': "Invalid RAKP4 integrity code (wrong Kg?)"})
             return
         self.sessionid = self.pendingsessionid
-        self.integrityalgo = 'sha1'
+        self.integrityalgo = self.attemptedhash
         self.confalgo = 'aes'
         self.sequencenumber = 1
         self.sessioncontext = 'ESTABLISHED'
@@ -1571,13 +1623,15 @@ class Session(object):
             # drop it and carry about our business.
             return
         if self.servermode:
-            self.seqlun = payload[4]
+            self.seqlun = payload[4] >> 2
             self.clientaddr = payload[3]
             self.clientnetfn = (payload[1] >> 2) + 1
+            self.rqlun = payload[4] & 0b11
             self.clientcommand = payload[5]
             self._parse_payload(payload)
             return
-        entry = (payload[1] >> 2, payload[4], payload[5])
+        # payload[4] is rqSeq + rsLUN, we only need the rqSeq
+        entry = (payload[1] >> 2, payload[4] >> 2, payload[5])
         if self._lookup_request_entry(entry):
             self._remove_request_entry(entry)
 
@@ -1613,8 +1667,8 @@ class Session(object):
         self.expectednetfn = 0x1ff
         self.expectedcmd = 0x1ff
         if not self.servermode:
-            self.seqlun += 4  # prepare seqlun for next transmit
-            self.seqlun &= 0xff  # when overflowing, wrap around
+            self.seqlun += 1  # prepare seqlun for next transmit
+            self.seqlun &= 0x3f  # when overflowing, wrap around
         with util.protect(WAITING_SESSIONS):
             Session.waiting_sessions.pop(self, None)
         # render retry mechanism utterly incapable of
@@ -1655,6 +1709,7 @@ class Session(object):
             self._mark_broken()
             return
         elif self.sessioncontext == 'FAILED':
+            self.lastpayload = None
             self.nowait = False
             return
         if self.sessioncontext == 'OPENSESSION':
@@ -1735,17 +1790,21 @@ class Session(object):
                     self.expiration = self.timeout + _monotonic_time()
                 Session.waiting_sessions[self]['timeout'] = self.expiration
 
-    def logout(self):
+    def logout(self, sessionok=True):
         if not self.logged:
             return {'success': True}
         if self.cleaningup:
             self.nowait = True
+        if self.sol_handler:
+            self.raw_command(netfn=6, command=0x49, data=(1, 1, 0, 0, 0, 0),
+                             retry=sessionok)
         self.raw_command(command=0x3c,
                          netfn=6,
                          data=struct.unpack("4B",
                                             struct.pack("I", self.sessionid)),
                          retry=False)
         # stop trying for a keepalive,
+        self.lastpayload = None
         with util.protect(KEEPALIVE_SESSIONS):
             Session.keepalive_sessions.pop(self, None)
         self.logged = 0
